// 008 u_c_basic_file
// 21 nov 2004

// -- (C) John COLIBRI 2006
// -- documentation: http://www.jcolibri.com

(*$r+*)

// -- add: concat, hex dump

unit u_c_basic_file;
  interface
    uses u_c_basic_object, u_c_file_name;

    type c_basic_file= class(c_basic_object)
                         private
                           m_file: File;
                         public
                           m_c_file_name: c_file_name;
                           m_io_result: Integer;
                           m_display_open: Boolean;
                           m_is_open: Boolean;

                           m_size: Integer;

                           m_write_begin, m_write_end: Integer;

                           constructor create_basic_file(p_name, p_complete_name: string); Virtual;

                           function f_check_and_display_io_result: Boolean;

                           function f_erase: Boolean;
                           function f_create: Boolean;
                           function f_force_create: Boolean; Virtual;
                           function f_rename_previous_and_create: Boolean;
                           function f_rename_previous_and_create_with_extension(p_extension: String): Boolean;
                           function f_open: Boolean; Virtual;
                           function f_open_or_create: Boolean;
                           function f_append: Boolean;
                           function f_exists: boolean;

                           procedure rename_file(p_new_name: String);
                           procedure force_rename_file(p_new_name: String);

                           procedure change_file_extension(p_extension: String);
                           procedure change_complete_name(p_complete_name: String);
                           function f_change_file_extension(p_extension: String): Boolean;

                           function f_test_or_create_u_path: String;

                           procedure seek_file(p_position: Integer); Virtual;
                           function f_seek_end_of_file: Integer; Virtual;
                           function f_file_position: Integer; Virtual;
                           function f_file_size: Integer; Virtual;
                           function f_byte_count: Integer;

                           function f_eof: Boolean;

                           function f_block_write(p_pt: Pointer; p_length: Integer): Boolean;
                           function f_block_read(p_pt: Pointer; p_length: Integer): Boolean;
                           procedure flush_file;

                           procedure copy_file(p_copy_name: String);

                           procedure truncate_file;
                           procedure close_file; Virtual;

                           Destructor Destroy; Override;
                       end;

    function f_io_error(p_error: Integer): String;

  implementation
    uses Classes, SysUtils, FileCtrl, u_loaded,
        u_c_display, u_dir;

    procedure start;
      begin
      end; // start

    function f_io_error(p_error: Integer): String;
      begin
        case p_error Of
          5 : Result:= 'ds: ReadOnly';
          100 : Result:= 'disk read error';
          101 : Result:= 'disk write error';
          102 : Result:= 'forgot Assign(file, ';
          103 : Result:= 'file not open';
          104 : Result:= 'file not open for reading';
          105 : Result:= 'file not open for writing';
          106 : Result:= 'invalid numeric format';
          else Result:= 'err_?? '+ IntToStr(p_error);
        end; // case
      end; // f_io_error

    function c_basic_file.f_check_and_display_io_result: Boolean;
      var l_result: Integer;
      begin
        l_result:= IoResult;
        if l_result<> 0
          then display('*** '+ m_c_file_name.m_complete_name+ ' '+ f_io_error(l_result));
        Result:= l_result= 0;
      end; // f_check_and_display_ioresult

    constructor c_basic_file.create_basic_file(p_name, p_complete_name: string);
      begin
        Inherited create_basic_object(p_name);

        m_c_file_name:= f_c_complete_file_name(p_name, p_complete_name);
        AssignFile(m_file, p_complete_name);
      end; // create_basic_file

    Destructor c_basic_file.Destroy;
      begin
        m_c_file_name.Free;
        close_file;
      end; // Destroy

    // -- open etc

    function c_basic_file.f_erase: Boolean;
      begin
        if m_is_open
          then begin
              display('*** already opened. Exit');
              Result:= False;
              exit;
            end;

        Result:= True;
        if f_exists
          then begin
              (*$i-*)
              erase(m_file);
              (*$i+*)
              Result:= f_check_and_display_io_result;
            end;
      end; // f_erase

    function c_basic_file.f_create: Boolean;
      begin
        if m_is_open
          then begin
              display('*** already opened. Exit');
              Result:= False;
              exit;
            end;

        if f_exists
          then Result:= False
          else begin
              Rewrite(m_file);
              Result:= True;
              m_is_open:= True;
            end;
      end; // f_create

    function c_basic_file.f_force_create: Boolean;
      begin
        // display('> f_force_create');
        if m_is_open
          then begin
              display('*** f_force_create: already opened. Exit');
              Result:= False;
              exit;
            end;

        f_erase;
        // display('f_force_create '+ m_c_file_name.m_complete_name);
        (*$i-*)
        Rewrite(m_file, 1);
        (*$i+*)

        Result:= f_check_and_display_io_result;

        m_size:= 0;
        if Result
          then m_is_open:= True;
        // display('< f_force_create');
      end; // f_force_create

    function c_basic_file.f_rename_previous_and_create: Boolean;
      var l_renamed_name, l_extension: String;
          l_file: File;
      begin
        if m_is_open
          then begin
              display('*** f_rename_previous_and_create: already opened. Exit');
              Result:= False;
              exit;
            end;

        if f_exists
          then begin
              display('  '+ m_c_file_name.m_complete_name+ ' exists: must rename old');
              with m_c_file_name do
              begin
                display_name_parts;
                if m_extension= ''
                  then l_extension:= '~ED'
                  else l_extension:= '~'+ m_extension;
              end; // with m_c_file_name

              with c_file_name.create_file_name('new', m_c_file_name.m_drive, m_c_file_name.m_path,
                m_c_file_name.m_name, l_extension) do
              begin
                display_name_parts;
                l_renamed_name:= m_complete_name;
                Free;
              end;

              display('changed name to >'+ l_renamed_name+ '<');

              // -- if this .~ED exists, erase it
              // -- ds: must not be in use !!
              if FileExists(l_renamed_name)
                then begin
                    display('  previous renamed exist: delete');
                    Assign(l_file, l_renamed_name);
                    Erase(l_file);
                  end;

              System.rename(m_file, l_renamed_name);
              display('  has renamed the disk file');

              // -- since System.rename has assigned, re-assign
              display('  re-assign '+ m_c_file_name.m_complete_name);
              Assign(m_file, m_c_file_name.m_complete_name);
            end;

        Rewrite(m_file, 1);
        Result:= True;
        m_size:= 0;
        m_is_open:= True;
        display('ok renamed_and_create');
      end; // f_rename_previous_and_create

    function c_basic_file.f_rename_previous_and_create_with_extension(p_extension: String): Boolean;
        // -- check that "NAME.EXT" does not exist (if exists, return False)
        // -- rename "NAME.ACT", "NAME.~ACT"
        // -- save as "NAME.EXT"
      var l_name_with_new_extension: String;
      begin
        if m_is_open
          then begin
              Result:= False;
              display('*** f_rename_previous_and_create_with_extension: already opened. Exit');
              exit;
            end;

        // -- rename
        with m_c_file_name do
          if UpperCase(m_extension)<> UpperCase(p_extension)
            then begin
                display('change the ext');
                l_name_with_new_extension:= ChangeFileExt(m_complete_name, '.'+ p_extension);
                if FileExists(l_name_with_new_extension)
                  then begin
                      Result:= False;
                      display('*** f_rename_previous_and_create_with_extension: NAME.NEW_EXT already exists. Exit');
                      exit;
                    end;

                rename_file(l_name_with_new_extension);
                change_extension(p_extension);
              end;

        Result:= f_rename_previous_and_create;
      end; // f_rename_previous_and_create_with_extension

    function c_basic_file.f_open: Boolean;
      begin
        if m_is_open
          then begin
              display('*** already opened. Exit');
              Result:= False;
              exit;
            end;

        if m_display_open
          then display('open '+ m_c_file_name.m_complete_name);
        (*$i-*)
        Reset(m_file, 1);
        (*$i+*)
        m_io_result:= IoResult;
        Result:= m_io_result= 0;
        if Result
          then begin
              m_size:= FileSize(m_file);
              m_is_open:= True;
            end
          else display('*** c_basic_file could not open '+ m_c_file_name.m_complete_name
            + ' '+ f_io_error(m_io_result)+ '<');
      end; // f_open

    function c_basic_file.f_open_or_create: Boolean;
      begin
        if f_exists
          then Result:= f_open                                                                        else Result:= f_force_create;
      end; // f_open_or_create

    function c_basic_file.f_exists: boolean;
      begin
        Result:= FileExists(m_c_file_name.m_complete_name);
      end; // f_exists

    function c_basic_file.f_append: Boolean;
      begin
        if f_open
          then begin
              seek_file(m_size);
              Result:= True;
            end
          else begin
              display('f_append: could not open ! ');
              Result:= False;
            end;
      end; // f_append

    procedure c_basic_file.rename_file(p_new_name: String);
      var l_complete_new_name: String;
      begin
        with m_c_file_name do
        begin
          l_complete_new_name:= m_drive+ ':'+ f_with_ending_slash(m_path)+ p_new_name;

          display('  rename '+ m_name+ '.'+ m_extension+ ' into '+ p_new_name);
          // -- close
          close_file;

          display('do rename');
          if FileExists(l_complete_new_name)
            then begin
                display('*** new exists');
                exit;
              end;

          System.rename(m_file, p_new_name);

          // -- update the file name
          change_file_and_extension_name(ExtractFileName(p_new_name));
        end; // wiht m_c_file_name
      end; // rename_file

    procedure c_basic_file.force_rename_file(p_new_name: String);
      var l_complete_new_name: String;
          l_file: File;
      begin
        with m_c_file_name do
        begin
          l_complete_new_name:= m_drive+ ':'+ f_with_ending_slash(m_path)+ p_new_name;

          display('  rename '+ m_name+ '.'+ m_extension+ ' into '+ p_new_name);
          // -- close
          close_file;

          display('do rename');
          if FileExists(l_complete_new_name)
            then begin
                display('erase previous '+ l_complete_new_name);
                Assign(l_file, l_complete_new_name);
                erase(l_file);
              end;

          System.rename(m_file, p_new_name);

          // -- update the file name
          change_file_and_extension_name(ExtractFileName(p_new_name));
        end; // wiht m_c_file_name
      end; // force_rename_file

    procedure c_basic_file.change_file_extension(p_extension: String);
        // -- only in memory
      var l_length: Integer;
      begin
        with m_c_file_name do
        begin
          l_length:= Length(p_extension);
          if l_length> 0
            then
              if p_extension[1]= '.'
                then delete(p_extension, 1, 1);
          m_extension:= p_extension;

          m_complete_name:= m_drive+ ':'
             + f_with_ending_slash(m_path)
             + m_name+ '.'+ m_extension;
          display('new name >'+ m_complete_name+ '<');
          Assign(m_file, m_complete_name);
        end; // with m_c_file_name
      end; // change_file_extension

    function c_basic_file.f_change_file_extension(p_extension: String): Boolean;
        // -- DISK+ memory.
        // -- if "NAME.NEW" exists, return False
      var l_name_with_new_extension: String;
      begin
        if m_is_open
          then begin
              Result:= False;
              display('*** f_change_file_extension: Open. Exit');
              exit;
            end;

        // -- rename
        with m_c_file_name do
          if UpperCase(m_extension)<> UpperCase(p_extension)
            then begin
                display('change the ext');
                l_name_with_new_extension:= ChangeFileExt(m_complete_name, '.'+ p_extension);
                if FileExists(l_name_with_new_extension)
                  then begin
                      Result:= False;
                      display('*** f_change_file_extension: NAME.NEW_EXT already exists. Exit');
                      exit;
                    end;

                rename_file(l_name_with_new_extension);
                change_extension(p_extension);
                Result:= True;
              end
            else Result:= False;
      end; // f_change_file_extension

    procedure c_basic_file.change_complete_name(p_complete_name: String);
      begin
        m_c_file_name.m_complete_name:= p_complete_name;
        m_c_file_name.decompose;

        Assign(m_file, m_c_file_name.m_complete_name);
      end; // change_complete_name

    function c_basic_file.f_test_or_create_u_path: String;
      begin
        with m_c_file_name do
        begin
          Result:= f_drive_and_path+ 'u_\';
          if Not DirectoryExists(Result)
            then ForceDirectories(Result);
        end;
        // display(l_result_name);
      end; // f_test_or_create_u_path

    // -- positioning

    procedure c_basic_file.seek_file(p_position: Integer);
      begin
        System.Seek(m_file, p_position);
      end; // seek_file

    function c_basic_file.f_seek_end_of_file: Integer;
      begin
        Result:= f_file_size;
        seek_file(Result);
        m_size:= Result;
      end; // f_seek_end_of_file

    function c_basic_file.f_byte_count: Integer;
      begin
        Result:= FileSize(m_file);
      end; // f_byte_count

    function c_basic_file.f_file_position: Integer;
      begin
        Result:= FilePos(m_file);
      end; // f_file_position

    function c_basic_file.f_file_size: Integer;
      begin
        Result:= FileSize(m_file);
      end; // f_file_size

    function c_basic_file.f_eof: Boolean;
      begin
        Result:= eof(m_file);
      end; // f_eof

    // -- reading and writing

    function c_basic_file.f_block_write(p_pt: Pointer; p_length: Integer): Boolean;
      var l_bytes_written: Integer;
      begin
        m_write_begin:= FilePos(m_file);
        BlockWrite(m_file, p_pt^, p_length, l_bytes_written);
        m_write_end:= FilePos(m_file);
        Result:= l_bytes_written= p_length;
        if not Result
          then begin
              display('*** BlockWrite '+ IntToStr(p_length)+ ' <> '+ IntToStr(l_bytes_written));
            end;
        m_size:= FileSize(m_file);
      end; // f_block_write

    function c_basic_file.f_block_read(p_pt: Pointer; p_length: Integer): Boolean;
      var l_bytes_read: Integer;
      begin
        BlockRead(m_file, p_pt^, p_length, l_bytes_read);
        Result:= l_bytes_read= p_length;
      end; // f_block_read

    procedure c_basic_file.flush_file;
      var l_position: Integer;
      begin
        l_position:= FilePos(m_file);
        close_file;
        f_open;
        Seek(m_file, l_position);
      end; // flush_file

    procedure c_basic_file.copy_file(p_copy_name: String);
      var fluxSource, fluxDestination: TFileStream;
      begin
        fluxSource:= TFileStream.Create(m_c_file_name.m_complete_name, fmOpenRead);
        try
          (*$i-*)
          fluxDestination:= TFileStream.Create(p_copy_name, fmOpenWrite or fmCreate);
          (*$i+*)
          f_check_and_display_io_result;
          try
            fluxDestination.CopyFrom(fluxSource, fluxSource.Size);
            display('copied '+ m_c_file_name.m_complete_name+ ' -> '+ p_copy_name);
          finally
            fluxDestination.Free;
          end;
        finally
          fluxSource.Free;
        end;
      end; // copy_file

    procedure c_basic_file.truncate_file;
      begin
        truncate(m_file);
      end; // truncate_file

    procedure c_basic_file.close_file;
      begin
        if m_is_open
          then begin
              System.Close(m_file);
              m_is_open:= False;
            end;

        m_size:= 0;
      end; // close_file

    begin // u_basic_file
      display_loaded('u_c_basic_file', @start);
    end. // u_basic_file


